/*******************************************Oled 6-page weather station******************************************************************************************************************
 * Software I used is partly Copyright (c) 2015 by Daniel Eichhorn,  http://blog.squix.ch
 * To make this sketch work install the libraries of Daniel from Github called son-streaming-parser-master.zip 
 * All the rest of the libs and code including the Adafruit DHTxx library is included locally  so the h and cpp files can be found within de sketch directory
 * Your local location can be searched on http://www.wunderground.com  like for example mine was http://www.wunderground.com/nl/maarssen
 * Your thingspeak API read key can be generated in your thinkgspeak channel that you want to display
 * 
 * My hardware consists of an ESP8266 version ESP-07. 
 * I also use a standard 128x64 OLED display (2.54" large with SSD1306 driver IC and an I2C base address 0x3C), a DHT22 combined Temperature/Humidity sensor and Neopixels from Adafruit to
 * color up the acrylic body of the weather station
 * 
 * 
 * Available GPIO on ESP-07: 1(blue led),2,3,4(SDA),5(SCL),12,13,14,16
 * 
 * IO connections currently used:
 * ESP-07 pin 4   -   to OLED SDA
 * ESP-07 pin 5   -   to OLED SCL
 * ESP-07 pin 12  -   DHT22 datapin (note: somehow pin16 does work on ESP-07 module (tested as PWM output) but does NOT want to control the DHT22)
 * ESP-07 pin 14  -   Neopixel RGB ledstrip datapin
 * 
 * Soldered on my ESP-07 (same as ESP-12) variant of the ESP8266 board:
 * Connect ESP8266 CH_PD and REST pulled up with 3k3 to 3.3V
 * Connect ESP8266 GPIO0 + GPIO2 pulled up with 3k3 to 3.3V
 * Connect ESP8266 GPIO15 to GROUND
 *  
 *  Software status: the Neopixel strip is defined as BRGW but actual sequence turns out to be WRBG: sort this out...
 * 
 * Current version  : 1.4
 * Revision history : 1.0 - initial revision with all libs/included h/cpp files local except ESP8266 and Json lib of Daniel Eichhorn
 *                    1.1 - Neopixels added
 *                    1.2 - Neopixel fade pattern added. Updating of the pattern is done using the Ticker functionality. update_my_neopixels.attach_ms(1,update_fading_of_my_neopixels)
 *                          in the function it calls (update_fading_of_my_neopixels) its also possible to change the color of the fading
 *                    1.3 - Added the neopixel color choice depending on current temp + forecast as a new ticker function, does this every x minutes
 *                    1.4 - Changed the Neopixel color choice algorithm, also added lightblue color to indicate it might freeze at night (this I did so I can protect my garden watertap!)
 *****************************************************************************************************************************************************************************************/


#include <ESP8266WiFi.h>
#include <Ticker.h>
#include <Adafruit_NeoPixel.h> //lib is installed, for easy RGB led control by using Adafruit's Neopixel ledstrips
#include <JsonListener.h>
#include <stdio.h>
#include <stdlib.h> //above 2 are so I can use abs() function
#include "SSD1306.h"
#include "SSD1306Ui.h"
#include "Wire.h"
#include "WundergroundClient.h"
#include "WeatherStationFonts.h"
#include "WeatherStationImages.h"
#include "TimeClient.h"
#include "ThingspeakClient.h"
#include "DHT.h"


#define baudrate 115200

// DHT Settings
#define DHTPIN 12     // what digital pin we're connected to. 

// Uncomment whatever type you're using!
//#define DHTTYPE DHT11   // DHT 11
#define DHTTYPE DHT22   // DHT 22  (AM2302), AM2321
//#define DHTTYPE DHT21   // DHT 21 (AM2301)

//Neopixel settings
#define NEOPIXELPIN     14 //esp8266 pin which is connected to the Neopixel strip
#define NUMPIXELS       12 //the number of Neopixels on the strip (I cut a strio of 12 neopixels and mounted these around the bezel of the OLED display I used)
#define TWENTY_MINUTES  1200 //check every 20 minutes the forecast and change the neopixel colors using Ticker + callback function
#define TEN_MINUTES     600
#define FIVE_MINUTES    300

/***************************
 * Begin Neopixel Patterns
 **************************/

// DVE added this to make use of Adafruit Neopixel Patterns code which includes Fading!
// Pattern types supported:
enum  pattern { FADE };
// Patern directions supported:
enum  direction { FORWARD, REVERSE };

// NeoPattern Class - derived from the Adafruit_NeoPixel class
class NeoPatterns : public Adafruit_NeoPixel
{
    public:
 
    // Member Variables:  
    pattern  ActivePattern;  // which pattern is running
    direction Direction;     // direction to run the pattern
    
    unsigned long Interval;   // milliseconds between updates
    unsigned long lastUpdate; // last update of position
    
    uint32_t Color1, Color2;  // What colors are in use
    uint16_t TotalSteps;  // total number of steps in the pattern
    uint16_t Index;  // current step within the pattern
    
    void (*OnComplete)();  // Callback on completion of pattern
    
    // Constructor - calls base-class constructor to initialize strip
    NeoPatterns(uint16_t pixels, uint8_t pin, uint8_t type, void (*callback)())
    :Adafruit_NeoPixel(pixels, pin, type)
    {
        OnComplete = callback;
    }
    
    // Update the pattern
    void Update()
    {
        if((millis() - lastUpdate) > Interval) // time to update
        {
            lastUpdate = millis();
            FadeUpdate();
        }
    }

    // Increment the Index and reset at the end
    void Increment()
    {
        if (Direction == FORWARD)
        {
           Index++;
           if (Index >= TotalSteps)
            {
                Index = 0;
                if (OnComplete != NULL)
                {
                    OnComplete(); // call the comlpetion callback
                }
            }
        }
        else // Direction == REVERSE
        {
            --Index;
            if (Index <= 0)
            {
                Index = TotalSteps-1;
                if (OnComplete != NULL)
                {
                    OnComplete(); // call the comlpetion callback
                }
            }
        }
    }
    
    // Reverse pattern direction
    void Reverse()
    {
        if (Direction == FORWARD)
        {
            Direction = REVERSE;
            Index = TotalSteps-1;
        }
        else
        {
            Direction = FORWARD;
            Index = 0;
        }
    }
    
    
    
    // Initialize for a Fade
    void Fade(uint32_t color1, uint32_t color2, uint16_t steps, uint8_t interval, direction dir = FORWARD)
    {
        ActivePattern = FADE;
        Interval = interval;
        TotalSteps = steps;
        Color1 = color1;
        Color2 = color2;
        Index = 0;
        Direction = dir;
    }

    void FadeUpdate()
    {
        // Calculate linear interpolation between Color1 and Color2
        // Optimise order of operations to minimize truncation error
        uint8_t red = ((Red(Color1) * (TotalSteps - Index)) + (Red(Color2) * Index)) / TotalSteps;
        uint8_t green = ((Green(Color1) * (TotalSteps - Index)) + (Green(Color2) * Index)) / TotalSteps;
        uint8_t blue = ((Blue(Color1) * (TotalSteps - Index)) + (Blue(Color2) * Index)) / TotalSteps;
        uint8_t white = ((White(Color1) * (TotalSteps - Index)) + (White(Color2) * Index)) / TotalSteps; //DVE added for RGBW strips
        
        ColorSet(Color(red, green, blue, white));//DVE added white for RGBW strip
        show();
        Increment();
    }

    // Calculate 50% dimmed version of a color (used by ScannerUpdate)
    uint32_t DimColor(uint32_t color)
    {
        // Shift R, G and B components one bit to the right
        uint32_t dimColor = Color(Red(color) >> 1, Green(color) >> 1, Blue(color) >> 1);
        return dimColor;
    }
 
    // Set all pixels to a color (synchronously)
    void ColorSet(uint32_t color)
    {
        for (int i = 0; i < numPixels(); i++)
        {
            setPixelColor(i, color);
        }
        show();
    }
 
    //DVE added this because I have a strip with white included so WRGB
    // Returns the White component of a 32-bit color
    uint8_t White(uint32_t color)
    {
        return (color >> 24) & 0xFF;
    }

    // Returns the Red component of a 32-bit color
    uint8_t Red(uint32_t color)
    {
        return (color >> 16) & 0xFF;
    }
 
    // Returns the Green component of a 32-bit color
    uint8_t Green(uint32_t color)
    {
        return (color >> 8) & 0xFF;
    }
 
    // Returns the Blue component of a 32-bit color
    uint8_t Blue(uint32_t color)
    {
        return color & 0xFF;
    }
    
    // Input a value 0 to 255 to get a color value.
    // The colours are a transition r - g - b - back to r.
    uint32_t Wheel(byte WheelPos)
    {
        WheelPos = 255 - WheelPos;
        if(WheelPos < 85)
        {
            return Color(255 - WheelPos * 3, 0, WheelPos * 3);
        }
        else if(WheelPos < 170)
        {
            WheelPos -= 85;
            return Color(0, WheelPos * 3, 255 - WheelPos * 3);
        }
        else
        {
            WheelPos -= 170;
            return Color(WheelPos * 3, 255 - WheelPos * 3, 0);
        }
    }
}; //end of class
     
//define a neopattern (nr of pixels, pixelpin, type)
NeoPatterns pixels(NUMPIXELS, NEOPIXELPIN, NEO_BRGW + NEO_KHZ800, &pixelsComplete);


//------------------------------------------------------------
//Completion Routines - get called on completion of a pattern
//------------------------------------------------------------
 
// pixels Completion Callback
void pixelsComplete()
{
  pixels.Reverse(); //reverses the fading direction from color1-->color2 to color2-->color1
}

/***************************
 * End Neopixel Patterns
 **************************/
 
//boolean toggle = 0; //for debugging using a led
 
/***************************
 * Begin Settings
 **************************/
// WIFI
const char* WIFI_SSID = "Kaka"; 
const char* WIFI_PWD = "EwenisWooley";

// Setup
const int UPDATE_INTERVAL_SECS = 10 * 60; // Update every 10 minutes

// Display Settings
const int I2C_DISPLAY_ADDRESS = 0x3c;
const int SDA_PIN = 4; //SDA and SCL pins on ESP-07 variant of ESP8266
const int SDC_PIN = 5;

// TimeClient settings
const float UTC_OFFSET = 1;

// Wunderground Settings
const boolean IS_METRIC = true;
const String WUNDERGRROUND_API_KEY = "ebedc5a1e35c2cf7";
const String WUNDERGROUND_COUNTRY = "NL";
String WUNDERGROUND_CITY = "MAARSSEN";

//Thingspeak Settings
const String THINGSPEAK_CHANNEL_ID = "50958";
const String THINGSPEAK_API_READ_KEY = "UGMS3TIO1VVB999A";

// Initialize the oled display for address 0x3c
SSD1306   display(I2C_DISPLAY_ADDRESS, SDA_PIN, SDC_PIN);
SSD1306Ui ui     ( &display );

/***************************
 * End Settings
 **************************/

TimeClient timeClient(UTC_OFFSET);

// Set to false, if you prefere imperial/inches, Fahrenheit
WundergroundClient wunderground(IS_METRIC);

ThingspeakClient thingspeak;


// Initialize the temperature/ humidity sensor
DHT dht(DHTPIN, DHTTYPE);


// this array keeps function pointers to all frames
// frames are the single views that slide from right to left
bool (*frames[])(SSD1306 *display, SSD1306UiState* state, int x, int y) = { drawFrame1, drawFrame2, drawFrame3, drawFrame4, drawFrame5, drawFrame6 };
int numberOfFrames = 6;

// flag changed in the ticker function every 10 minutes
bool readyForWeatherUpdate = false;

String lastUpdate = "--";

float humidity = 0.0;
float temperature = 0.0;

Ticker ticker;
Ticker update_my_neopixels; //DVEDEBUG experiment: can I use this to update my neopixels fast enough to make the fading look good BUT: without interfering with the oled display frame changes?
Ticker change_neopix_color;

void setup() 
{
  Serial.begin(baudrate);
  Serial.println();
  Serial.println();
  
  // initialize display
  display.init();
  display.clear();
  display.display();

  display.flipScreenVertically(); //DVE added this to rotate the view on the OLED display 180 degrees
  display.setFont(ArialMT_Plain_10);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.setContrast(255);

  WiFi.begin(WIFI_SSID, WIFI_PWD);
  
  int counter = 0;
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
    display.clear();
    display.drawString(64, 10, "Connecting to WiFi");
    display.drawXbm(46, 30, 8, 8, counter % 3 == 0 ? activeSymbole : inactiveSymbole);
    display.drawXbm(60, 30, 8, 8, counter % 3 == 1 ? activeSymbole : inactiveSymbole);
    display.drawXbm(74, 30, 8, 8, counter % 3 == 2 ? activeSymbole : inactiveSymbole);
    display.display();
    
    counter++;
  }
  
  ui.setTargetFPS(30); //if FPS (frames per second) is made smaller the windows change faster on the OLED display

  ui.setActiveSymbole(activeSymbole);
  ui.setInactiveSymbole(inactiveSymbole);

  // You can change this to
  // TOP, LEFT, BOTTOM, RIGHT
  ui.setIndicatorPosition(BOTTOM);

  // Defines where the first frame is located in the bar.
  ui.setIndicatorDirection(LEFT_RIGHT);

  // You can change the transition that is used
  // SLIDE_LEFT, SLIDE_RIGHT, SLIDE_TOP, SLIDE_DOWN
  ui.setFrameAnimation(SLIDE_LEFT);

  // Add frames
  ui.setFrames(frames, numberOfFrames);

  // Inital UI takes care of initalising the display too.
  ui.init();

  Serial.println("");

  display.flipScreenVertically(); //DVE once more seems to be required: rotate the view on the OLED display 180 degrees

  updateData(&display);

  ticker.attach(UPDATE_INTERVAL_SECS, setReadyForWeatherUpdate);
  update_my_neopixels.attach_ms(1,update_fading_of_my_neopixels); //using the ESP8266 Ticker function to call the neopixel pattern fading every 1mS, interferes slightly with the frames changing but not too annoying
  change_neopix_color.attach(TWENTY_MINUTES,change_neopixel_color); //this function checks each X minutes the forecast like max temp, compares to current temp and changes the neopixel color: red means its getting hotter, blue = colder

  pixels.begin(); // This initializes the NeoPixel library.
  //Next one is IMPORTANT CALL! It triggers Fade() once which is required or the station crashes. Cannot explain why but when not doing this the station crashes into an endless restart loop
  change_neopixel_color(); //Function runs via Ticker but call the function 1 time right away otherwise it takes x minutes before the color is updated to the right one
}

void loop() 
{  
  if (readyForWeatherUpdate && ui.getUiState().frameState == FIXED) {
    updateData(&display);
  }

  int remainingTimeBudget = ui.update();

  if (remainingTimeBudget > 0) 
  {
    // You can do some work here, don't do stuff if you are below your time budget.
  }
}

//DVE this function is called every 1ms using the Ticker option and is solely to be able to update the fading pattern of my neopixels. 
void update_fading_of_my_neopixels(void)
{
  pixels.Update();
}


//change the color of the neopixels every X minutes and depending on the weather forecast (index 0 = today, 2 = tomorrow, 4 is after tomorrow)
void change_neopixel_color(void)
{  pixels.Fade(0x00,0x00ADFFE6,2000,1,FORWARD);//light blue
  
  String todayhightemp = wunderground.getForecastHighTemp(0); //store the forecast today expected maximum temperature
  String todaylowtemp = wunderground.getForecastLowTemp(0); //
  String tomorrowhightemp = wunderground.getForecastHighTemp(2); //hightemp tomorrow
  String tomorrowlowtemp = wunderground.getForecastLowTemp(2); //and lowtemp as well
  String currtemp = wunderground.getCurrentTemp(); //also store the current temperature

  int todayhightemperature = todayhightemp.toInt();
  int todaylowtemperature = todaylowtemp.toInt();
  int tomorrowhightemperature = tomorrowhightemp.toInt();
  int tomorrowlowtemperature = tomorrowlowtemp.toInt();
  int currenttemperature = currtemp.toInt();

  //calculate the delta temps between forecast today's min+max and tomorrow's min+max
  int delta_max_temp = abs(tomorrowhightemperature - todayhightemperature);
  int delta_min_temp = abs(tomorrowlowtemperature - todaylowtemperature);
  
  //change fading color depending on the current + forecast temp
  if (tomorrowlowtemperature <= 0) //to warn me if it will freeze at night (so I can go out to my garden and protect my water tap :)  )
    pixels.Fade(0x00,0xFF000000,2000,1,FORWARD);//light blue
  else if( (tomorrowhightemperature > (todayhightemperature + 2)) && (tomorrowlowtemperature > (todaylowtemperature + 2)) )//if tomorrow the weather will be warmer
    pixels.Fade(0x00,0xFF0000,2000,1,FORWARD);//change to red
  else if(delta_max_temp <= 2 && delta_min_temp <= 2) //there will be no significant difference in temperatures
    pixels.Fade(0x00,0xFF,2000,1,FORWARD);//green
  else 
    pixels.Fade(0x00,0xFF00,2000,1,FORWARD);//change color to blue because it will be colder tomorrow 
} 

void updateData(SSD1306 *display) {
  drawProgress(display, 10, "Updating time...");
  timeClient.updateTime();
  
  drawProgress(display, 30, "Updating conditions...");
  wunderground.updateConditions(WUNDERGRROUND_API_KEY, WUNDERGROUND_COUNTRY, WUNDERGROUND_CITY);
  
  drawProgress(display, 50, "Updating forecasts...");
  wunderground.updateForecast(WUNDERGRROUND_API_KEY, WUNDERGROUND_COUNTRY, WUNDERGROUND_CITY);
  
  drawProgress(display, 80, "Updating thingspeak...");
  thingspeak.getLastChannelItem(THINGSPEAK_CHANNEL_ID, THINGSPEAK_API_READ_KEY);
  lastUpdate = timeClient.getFormattedTime();
  
  drawProgress(display, 90, "Updating local temperature + humidity");
  humidity = dht.readHumidity();
  // Read temperature as Fahrenheit (isFahrenheit = true)
  temperature = dht.readTemperature(!IS_METRIC);
  drawProgress(display, 100, "Done...");
  readyForWeatherUpdate = false;
  delay(1000);
}

void drawProgress(SSD1306 *display, int percentage, String label) {
  display->clear();
  display->setTextAlignment(TEXT_ALIGN_CENTER);
  display->setFont(ArialMT_Plain_10);
  display->drawString(64, 10, label);
  display->drawRect(10, 28, 108, 12);
  display->fillRect(12, 30, 104 * percentage / 100 , 9);
  display->display();
}


bool drawFrame1(SSD1306 *display, SSD1306UiState* state, int x, int y) {
  display->setTextAlignment(TEXT_ALIGN_CENTER);
  display->setFont(ArialMT_Plain_10);
  String date = wunderground.getDate();
  int textWidth = display->getStringWidth(date);
  display->drawString(64 + x, 10 + y, date);
  display->setFont(ArialMT_Plain_24);
  String time = timeClient.getFormattedTime();
  textWidth = display->getStringWidth(time);
  display->drawString(64 + x, 20 + y, time);
  display->setTextAlignment(TEXT_ALIGN_LEFT);
}

bool drawFrame2(SSD1306 *display, SSD1306UiState* state, int x, int y) {
  display->setFont(ArialMT_Plain_10);
  display->drawString(60 + x, 10 + y, wunderground.getWeatherText());

  display->setFont(ArialMT_Plain_24);
  String temp = wunderground.getCurrentTemp() + "°C";
  display->drawString(60 + x, 20 + y, temp);
  int tempWidth = display->getStringWidth(temp);

  display->setFont(Meteocons_0_42);
  String weatherIcon = wunderground.getTodayIcon();
  int weatherIconWidth = display->getStringWidth(weatherIcon);
  display->drawString(32 + x - weatherIconWidth / 2, 10 + y, weatherIcon);
}

bool drawFrame3(SSD1306 *display, SSD1306UiState* state, int x, int y) {
  display->setTextAlignment(TEXT_ALIGN_CENTER);
  display->setFont(ArialMT_Plain_10);
  display->drawString(32 + x, 0 + y, "Humidity");
  display->drawString(96 + x, 0 + y, "Pressure");
  display->drawString(32 + x, 28 + y, "Precipit.");

  display->setFont(ArialMT_Plain_16);
  display->drawString(32 + x, 10 + y, wunderground.getHumidity());
  display->drawString(96 + x, 10 + y, wunderground.getPressure());
  display->drawString(32 + x, 38 + y, wunderground.getPrecipitationToday());
}

bool drawFrame4(SSD1306 *display, SSD1306UiState* state, int x, int y) {
  drawForecast(display, x, y, 0);
  drawForecast(display, x + 44, y, 2);
  drawForecast(display, x + 88, y, 4);
}

bool drawFrame5(SSD1306 *display, SSD1306UiState* state, int x, int y) {
  display->setTextAlignment(TEXT_ALIGN_CENTER);
  display->setFont(ArialMT_Plain_10);
  display->drawString(64 + x, 0 + y, "Outdoor");
  display->setFont(ArialMT_Plain_24);
  display->drawString(64 + x, 10 + y, thingspeak.getFieldValue(0) + "°C");
  display->drawString(64 + x, 30 + y, thingspeak.getFieldValue(1) + "mb");
}

bool drawFrame6(SSD1306 *display, SSD1306UiState* state, int x, int y) {
  display->setTextAlignment(TEXT_ALIGN_CENTER);
  display->setFont(ArialMT_Plain_10);
  display->drawString(64 + x, 0 + y, "Indoor");
  display->setFont(ArialMT_Plain_24);
  display->drawString(64 + x, 10 + y, String(temperature) + "°C");
  display->drawString(64 + x, 30 + y, String(humidity) + "%");
}

void drawForecast(SSD1306 *display, int x, int y, int dayIndex) {
  display->setTextAlignment(TEXT_ALIGN_CENTER);
  display->setFont(ArialMT_Plain_10);
  String day = wunderground.getForecastTitle(dayIndex).substring(0, 3);
  day.toUpperCase();
  display->drawString(x + 20, y, day);
  
  display->setFont(Meteocons_0_21);
  display->drawString(x + 20, y + 15, wunderground.getForecastIcon(dayIndex));

  display->setFont(ArialMT_Plain_16);
  display->drawString(x + 20, y + 37, wunderground.getForecastLowTemp(dayIndex) + "/" + wunderground.getForecastHighTemp(dayIndex));
  //display.drawString(x + 20, y + 51, );
  display->setTextAlignment(TEXT_ALIGN_LEFT);
}

void setReadyForWeatherUpdate() {
  Serial.println("Setting readyForUpdate to true");
  readyForWeatherUpdate = true;
}


